import numpy as np

from stimupy.components import image_base
from stimupy.components.angulars import wedge
from stimupy.components.radials import annulus, disc, ring
from stimupy.utils import resolution

__all__ = [
    "rectangle",
    "triangle",
    "cross",
    "parallelogram",
    "ellipse",
    "circle",
    "wedge",
    "annulus",
    "disc",
    "ring",
]


def rectangle(
    visual_size=None,
    ppd=None,
    shape=None,
    rectangle_size=None,
    rectangle_position=None,
    intensity_rectangle=1.0,
    intensity_background=0.0,
    rotation=0.0,
):
    """Draw a rectangle

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    rectangle_size : Number, Sequence[Number, Number]
        rectangle size [height, width], in degrees visual angle
    rectangle_position : Number, Sequence[Number, Number], or None (default)
        position of the rectangle, in degrees visual angle.
        If None, rectangle will be placed in center of image.
    intensity_rectangle : float, optional
        intensity value for rectangle, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "rectangle_mask"),
        and additional keys containing stimulus parameters
    """
    if rectangle_size is None:
        raise ValueError("rectangle() missing argument 'rectangle_size' which is not 'None'")

    # Resolve resolutions and get distances
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin="center",
    )
    xx = base["horizontal"]
    yy = base["vertical"]
    theta = np.deg2rad(rotation)
    rectangle_size = resolution.validate_visual_size(visual_size=rectangle_size)

    # Determine center position
    rect_posy = (base["visual_size"].height / 2) - (rectangle_size.height / 2)
    rect_posx = (base["visual_size"].width / 2) - (rectangle_size.width / 2)
    center_pos = (rect_posy, rect_posx)

    if rectangle_position is None:
        # If no position is given, place rectangle centrally
        rectangle_position = center_pos

    # Positions should always be positive
    rectangle_position = np.array(rectangle_position).clip(min=0)
    center_pos = np.array(center_pos).clip(min=0)

    # Determine shift
    rect_pos = resolution.shape_from_visual_size_ppd(rectangle_position, base["ppd"])
    center_pos = resolution.shape_from_visual_size_ppd(center_pos, base["ppd"])
    rect_shift = (np.array(rect_pos) - np.array(center_pos)).astype(int)

    # Rotate coordinate systems
    x = np.round(np.cos(theta) * xx - np.sin(theta) * yy, 8)
    y = np.round(np.sin(theta) * xx + np.cos(theta) * yy, 8)

    # Rounding for more robust behavior:
    x = np.round(x * (base["ppd"][0] * 2)) / (base["ppd"][0] * 2)
    y = np.round(y * (base["ppd"][0] * 2)) / (base["ppd"][0] * 2)

    # Draw rectangle
    img1 = np.where(x < rectangle_size.width / 2, 1, 0)
    img2 = np.where(x >= -rectangle_size.width / 2, 1, 0)
    img3 = np.where(y < rectangle_size.height / 2, 1, 0)
    img4 = np.where(y >= -rectangle_size.height / 2, 1, 0)
    img = img1 * img2 * img3 * img4

    # Shift rectangle
    img = np.roll(img, (rect_shift[0], rect_shift[1]), axis=(0, 1))

    # Does the rectangle fit?
    x1 = rectangle_size[1] / 2 * np.cos(theta)
    x2 = rectangle_size[1] / 2 * np.sin(theta)
    y1 = rectangle_size[0] / 2 * np.cos(theta)
    y2 = rectangle_size[0] / 2 * np.sin(theta)
    cy = x2 + y1 + np.abs(rect_shift[0] / base["ppd"][0])
    cy = np.floor(cy * base["ppd"][0]) / base["ppd"][0]
    cx = x1 + y2 + np.abs(rect_shift[1] / base["ppd"][1])
    cx = np.floor(cx * base["ppd"][1]) / base["ppd"][1]

    if (cy > base["visual_size"][0] / 2) or (cx > base["visual_size"][1] / 2):
        raise ValueError("stimulus does not fully fit into requested size")

    return {
        "img": img * (intensity_rectangle - intensity_background) + intensity_background,
        "rectangle_mask": img.astype(int),
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "shape": base["shape"],
        "rectangle_size": rectangle_size,
        "rectangle_position": rectangle_position,
        "intensity_background": intensity_background,
        "intensity_rectangle": intensity_rectangle,
        "rotation": rotation,
    }


def triangle(
    visual_size=None,
    ppd=None,
    shape=None,
    triangle_size=None,
    intensity_triangle=1.0,
    intensity_background=0.0,
    include_corners=True,
    rotation=0.0,
):
    """Draw a triangle

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    triangle_size : Number, Sequence[Number, Number]
        triangle size [height width], in degrees visual angle
    intensity_triangle : float, optional
        intensity value for triangle, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "triangle_mask"),
        and additional keys containing stimulus parameters
    """
    if triangle_size is None:
        raise ValueError("triangle() missing argument 'triangle_size' which is not 'None'")

    # Resolve resolutions and get distances
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin="center",
    )
    xx = base["horizontal"]
    yy = base["vertical"]
    triangle_size = resolution.validate_visual_size(visual_size=triangle_size)

    angle_diagonal = np.arctan(triangle_size[1] / triangle_size[0])
    angle_diagonal = np.rad2deg(angle_diagonal)
    theta = np.deg2rad(rotation + angle_diagonal)
    x = np.round(np.cos(theta) * xx - np.sin(theta) * yy, 8)

    # Split image in two parts following the diagonal
    if include_corners:
        img = np.where(x <= 0, 1, 0)
    else:
        fac = base["ppd"][0] / 2
        x = np.round(x * fac) / fac
        img = np.where(x < 0, 1, 0)

    # Create rectangular mask
    rect = rectangle(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rectangle_size=triangle_size,
        rotation=rotation,
    )
    img = img * rect["rectangle_mask"]

    return {
        "img": img * (intensity_triangle - intensity_background) + intensity_background,
        "triangle_mask": img.astype(int),
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "shape": base["shape"],
        "triangle_size": triangle_size,
        "intensity_background": intensity_background,
        "intensity_triangle": intensity_triangle,
        "rotation": rotation,
        "include_corners": include_corners,
    }


def cross(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_size=None,
    cross_thickness=None,
    cross_arm_ratios=(1.0, 1.0),
    intensity_cross=1.0,
    intensity_background=0.0,
    rotation=0.0,
):
    """Draw a cross

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    cross_size : Number, Sequence[Number, Number]
        cross size [height, width], in degrees visual angle
    cross_thickness : Number, Sequence[Number, Number]
        thickness of cross in degrees visual angle
    cross_arm_ratios : float or (float, float)
        ratio used to create arms (up-down, left-right)
    intensity_cross: float, optional
        intensity value for cross, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "cross_mask"),
        and additional keys containing stimulus parameters
    """
    if cross_size is None:
        raise ValueError("cross() missing argument 'cross_size' which is not 'None'")
    if cross_thickness is None:
        raise ValueError("cross() missing argument 'cross_thickness' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    cross_size = resolution.validate_visual_size(cross_size)
    cross_thickness = resolution.validate_visual_size(cross_thickness)

    if isinstance(cross_arm_ratios, (float, int)):
        cross_arm_ratios = (cross_arm_ratios, cross_arm_ratios)

    # Determine coordinate center
    cy = visual_size.height / 2
    cx = visual_size.width / 2
    theta = np.deg2rad(rotation)

    # Calculate cross placement based on ratios of cross legs
    updown = cross_size.height - cross_thickness[0]
    down = updown / (cross_arm_ratios[0] + 1)
    up = updown - down
    leftright = cross_size.width - cross_thickness[1]
    right = leftright / (cross_arm_ratios[1] + 1)
    left = leftright - right

    posy1 = cy - cross_size[0] / 2 + (down - up) * np.cos(theta) / 2
    posx1 = cx - cross_thickness[0] / 2 + (down - up) * np.sin(theta) / 2

    posy2 = cy - cross_thickness[1] / 2 + (right - left) * np.sin(-theta) / 2
    posx2 = cx - cross_size[1] / 2 + (right - left) * np.cos(-theta) / 2

    # Create cross as two rectangles
    rect1 = rectangle(
        visual_size=visual_size,
        ppd=ppd,
        rectangle_size=(cross_size[0], cross_thickness[0]),
        rectangle_position=(posy1, posx1),
        rotation=rotation,
    )

    rect2 = rectangle(
        visual_size=visual_size,
        ppd=ppd,
        rectangle_size=(cross_thickness[1], cross_size[1]),
        rectangle_position=(posy2, posx2),
        rotation=rotation,
    )

    img = rect1["img"] + rect2["img"]
    img[img > 1] = 1

    return {
        "img": img * (intensity_cross - intensity_background) + intensity_background,
        "cross_mask": img.astype(int),
        "shape": shape,
        "visual_size": visual_size,
        "ppd": ppd,
        "cross_size": cross_size,
        "cross_arm_ratios": cross_arm_ratios,
        "cross_thickness": cross_thickness,
        "intensity_background": intensity_background,
        "intensity_cross": intensity_cross,
        "rotation": rotation,
    }


def parallelogram(
    visual_size=None,
    ppd=None,
    shape=None,
    parallelogram_size=None,
    intensity_parallelogram=1.0,
    intensity_background=0.0,
    rotation=0.0,
):
    """Draw a parallelogram

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    parallelogram_size : [Number, Number, Number], [Number, Number], Number or None (default)
        parallelogram size [height, width, depth], in degrees visual angle
    intensity_parallelogram : float, optional
        intensity value for parallelogram, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "parallelogram_mask"),
        and additional keys containing stimulus parameters
    """
    if parallelogram_size is None:
        raise ValueError(
            "parallelogram() missing argument 'parallelogram_size' which is not 'None'"
        )
    if isinstance(parallelogram_size, (float, int)):
        parallelogram_size = (parallelogram_size, parallelogram_size, 0)
    if len(parallelogram_size) == 2:
        parallelogram_size = tuple(
            list(parallelogram_size)
            + [
                0,
            ]
        )

    # Resolve resolutions and get distances
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin="center",
    )
    xx = base["horizontal"]
    yy = base["vertical"]

    # Create rectangule
    rectangle_size = (parallelogram_size[0], parallelogram_size[1] + np.abs(parallelogram_size[2]))
    rect = rectangle(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rectangle_size=rectangle_size,
        rotation=rotation,
    )
    img = rect["img"]

    if parallelogram_size[2] != 0:
        if parallelogram_size[2] > 0:
            triangle_size = (parallelogram_size[0], np.abs(parallelogram_size[2]))
            rot1 = rotation
        else:
            triangle_size = (np.abs(parallelogram_size[2]), parallelogram_size[0])
            rot1 = rotation - 90

        angle_diagonal = np.arctan(triangle_size[1] / triangle_size[0])
        angle_diagonal = np.rad2deg(angle_diagonal)
        theta = np.deg2rad(rot1 + angle_diagonal)
        x = np.round(np.cos(theta) * xx - np.sin(theta) * yy, 8)

        # Shift diagonals so that resulting triangles cover corners of rectangle
        theta = np.deg2rad(rotation)
        pwidth = parallelogram_size[1] / 2 * base["ppd"][0]
        shift1 = int(np.round(pwidth) * np.sin(theta))
        shift2 = int(np.floor(pwidth) * np.cos(theta))

        # Split image in two parts following the diagonal
        tri1 = np.where(np.roll(x, (shift1, -shift2), axis=(0, 1)) < 0, 0, 1)
        tri1 = np.where(x < 0, tri1, 0)
        tri2 = np.where(np.roll(x, (-shift1, shift2), axis=(0, 1)) > 0, 0, 1)
        tri2 = np.where(x >= 0, tri2, 0)

        # Combine everything
        img = tri1 * img + tri2 * img

    return {
        "img": img * (intensity_parallelogram - intensity_background) + intensity_background,
        "parallelogram_mask": img.astype(int),
        "shape": base["shape"],
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "parallelogram_size": parallelogram_size,
        "intensity_background": intensity_background,
        "intensity_parallelogram": intensity_parallelogram,
        "rotation": rotation,
    }


def ellipse(
    visual_size=None,
    ppd=None,
    shape=None,
    radius=None,
    intensity_ellipse=1.0,
    intensity_background=0.0,
    rotation=0.0,
    origin="mean",
    restrict_size=True,
):
    """Draw an ellipse

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radius : Sequence[Number, Number], Number or None (default)
        ellipse radius [ry, rx] in degrees visual angle
    intensity_ellipse : float, optional
        intensity value for ellipse, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    restrict_size : Bool
        if False, allow ellipse to reach beyond image size (default: True)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "ellipse_mask"),
        and additional keys containing stimulus parameters
    """
    if radius is None:
        raise ValueError("ellipse() missing argument 'radius' which is not 'None'")

    # Resolve resolutions and get distances
    radius = resolution.validate_visual_size(visual_size=radius)
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin=origin,
    )

    xx = base["horizontal"]
    yy = base["vertical"]

    # Rotate coordinate systems
    theta = np.deg2rad(rotation)
    x = np.round(np.cos(theta) * yy - np.sin(theta) * xx, 8)
    y = np.round(np.sin(theta) * yy + np.cos(theta) * xx, 8)

    # Draw ellipse
    arr = np.sqrt(x**2 + (y * radius[0] / radius[1]) ** 2)
    img = np.where(arr <= radius[0], 1, 0)

    # Does ellipse fit?
    x1 = radius[1] * np.cos(theta)
    x2 = radius[1] * np.sin(theta)
    y1 = radius[0] * np.cos(theta)
    y2 = radius[0] * np.sin(theta)
    cy = np.floor((x2 + y1) * base["ppd"][0]) / base["ppd"][0]
    cx = np.floor((x1 + y2) * base["ppd"][1]) / base["ppd"][1]

    if restrict_size and ((cy > base["visual_size"][0] / 2) or (cx > base["visual_size"][1] / 2)):
        raise ValueError("stimulus does not fully fit into requested size")

    return {
        "img": img * (intensity_ellipse - intensity_background) + intensity_background,
        "ellipse_mask": img.astype(int),
        "shape": base["shape"],
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "radius": radius,
        "intensity_background": intensity_background,
        "intensity_ellipse": intensity_ellipse,
        "rotation": rotation,
    }


def circle(
    visual_size=None,
    ppd=None,
    shape=None,
    radius=None,
    intensity_circle=1.0,
    intensity_background=0.0,
    origin="mean",
    restrict_size=True,
):
    """Draw an ellipse

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radius : Number or None (default)
        circle radius in degrees visual angle
    intensity_circle : float, optional
        intensity value for circle, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    restrict_size : Bool
        if False, allow circle to reach beyond image size (default: True)

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the shape (key: "circle_mask"),
        and additional keys containing stimulus parameters
    """
    if radius is None:
        raise ValueError("circle() missing argument 'radius' which is not 'None'")
    if not isinstance(radius, (int, float)):
        raise ValueError("radius should be a single number")

    stim = ellipse(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radius=radius,
        intensity_ellipse=intensity_circle,
        intensity_background=intensity_background,
        rotation=0.0,
        origin=origin,
        restrict_size=restrict_size,
    )
    stim["circle_mask"] = stim["ellipse_mask"]
    stim["intensity_circle"] = intensity_circle
    stim.pop("ellipse_mask", "intensity_ellipse")
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 20,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "shapes_rectangle": rectangle(**default_params, rectangle_size=(4, 2.5)),
        "shapes_triangle": triangle(**default_params, triangle_size=(4, 2.5)),
        "shapes_cross": cross(**default_params, cross_size=(4, 2.5), cross_thickness=1, cross_arm_ratios=(1, 1)),
        "shapes_parallelogram": parallelogram(**default_params, parallelogram_size=(5.2, 3.1, 0.9)),
        "shapes_ellipse": ellipse(**default_params, radius=(4, 3)),
        "shapes_circle": circle(**default_params, radius=3),
        "shapes_disc": disc(**default_params, radius=3),
        "shapes_ring": ring(**default_params, radii=(1, 2)),
        "shapes_annulus": annulus(**default_params, radii=(1, 2)),
        "shapes_wedge": wedge(**default_params, angle=30, radius=4),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Tests whether validation of resolution components works as expected

Separately tests each of three components: visual size, shape, ppd.
For each component, the validation routine has two expected major behaviors:
- if input is valid, return canonical format (2-(named)tuple)
  - casts values to correct type (float or int)
  - expands 1D to 2D, i.e., if 1 value is provided, uses it for both dimensions
- if input is not valid, raise specific exceptions:
  - ValueError if input cannot be cast
  - ValueError if values are not in accepted ranges, i.e., not positive
  - TypeError if input is not a 1-Sequence or 2-Sequence

The tests do this through parameterization:
pytest runs each test for each of the specified parameter sets.

For the valid-input tests, each parameter set specifies the (valid) input,
and value (in canonical form) that should be returned.
If the actual returned value matches this expected return value, the test passes.

For the invalid-input tests, each parameter set specifies the (invalid) input,
and the Exception that should be raised.
If the function under testing raises this specified exception, the test passes.
"""

import pytest

from stimupy.utils import resolution


#############################
#    Validate visual size   #
#############################
@pytest.mark.parametrize(
    "visual_size, expected",
    [
        ((32, 32), (32, 32)),
        ((16, 32), (16, 32)),
        ((22.5, 48.1), (22.5, 48.1)),
        ((16), (16, 16)),
        ("32", (32, 32)),
        ((16, "32"), (16, 32)),
        (("16", "32"), (16, 32)),
        ((16.1, None), (16.1, None)),
        ((None, "32"), (None, 32)),
        ((None, None), (None, None)),
        (None, (None, None)),
    ],
)
def test_valid_visual_size(visual_size, expected):
    out = resolution.validate_visual_size(visual_size)
    assert out.width == expected[1] and out.height == expected[0]


@pytest.mark.parametrize(
    "visual_size, exception",
    [
        ((), TypeError),
        ("bla", ValueError),
        ((32, 32, 32), TypeError),
        ({32, 32}, TypeError),
        ((-32, 16), ValueError),
        ((16, -1), ValueError),
    ],
)
def test_invalid_visual_size(visual_size, exception):
    with pytest.raises(exception):
        resolution.validate_visual_size(visual_size)


#############################
#       Validate shape      #
#############################
@pytest.mark.parametrize(
    "shape, expected",
    [
        ((32, 32), (32, 32)),
        ((16, 32), (16, 32)),
        ((22.5, 48.1), (22, 48)),
        ((16), (16, 16)),
        ("32", (32, 32)),
        ((16, "32"), (16, 32)),
        (("16", "32"), (16, 32)),
        ((16.1, None), (16, None)),
        ((None, "32"), (None, 32)),
        ((None, None), (None, None)),
        (None, (None, None)),
    ],
)
def test_valid_shape(shape, expected):
    out = resolution.validate_shape(shape)
    assert out.width == expected[1] and out.height == expected[0]


@pytest.mark.parametrize(
    "shape, exception",
    [
        ((), TypeError),
        ("bla", ValueError),
        ((32, 32, 32), TypeError),
        ({32, 32}, TypeError),
        ((-32, 16), ValueError),
        ((16, 0), ValueError),
    ],
)
def test_invalid_shape(shape, exception):
    with pytest.raises(exception):
        resolution.validate_shape(shape)


############################
#       Validate ppd       #
############################
@pytest.mark.parametrize(
    "ppd, expected",
    [
        ((32, 32), (32, 32)),
        ((16, 32), (16, 32)),
        ((22.5, 48.1), (22.5, 48.1)),
        ((16), (16, 16)),
        ("32", (32, 32)),
        ((16, "32"), (16, 32)),
        (("16", "32"), (16, 32)),
        ((16.1, None), (16.1, None)),
        ((None, "32"), (None, 32)),
        ((None, None), (None, None)),
        (None, (None, None)),
    ],
)
def test_valid_ppd(ppd, expected):
    out = resolution.validate_ppd(ppd)
    assert out.horizontal == expected[1] and out.vertical == expected[0]


@pytest.mark.parametrize(
    "ppd, exception",
    [
        ((), TypeError),
        ("bla", ValueError),
        ((32, 32, 32), TypeError),
        ({32, 32}, TypeError),
        ((-32, 16), ValueError),
        ((16, 0), ValueError),
    ],
)
def test_invalid_ppd(ppd, exception):
    with pytest.raises(exception):
        resolution.validate_ppd(ppd)



import itertools

import numpy as np

from stimupy.components.shapes import parallelogram
from stimupy.utils import resolution

__all__ = [
    "mondrian",
    "corrugated_mondrian",
]


def mondrian(
    visual_size=None,
    ppd=None,
    shape=None,
    positions=None,
    sizes=None,
    intensities=None,
    intensity_background=0.5,
):
    """Draw Mondrian of given size and intensity at given position

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    positions : Sequence[tuple, ... ] or None (default)
        position (y, x) of each Mondrian in degrees visual angle
    sizes : Sequence[tuple, ... ] or None (default)
        size (height, width, depth) of Mondrian parallelograms in degrees visual angle;
        if only one number is given, squares will be drawn
    intensities : Sequence[Number, ... ] or None (default)
        intensity values of each Mondrian, if only one number is given
        all will have the same intensity
    intensity_background : float
        intensity value of background

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each Mondrian (key: "mondrian_mask"),
        and additional keys containing stimulus parameters
    """
    if positions is None:
        raise ValueError("mondrians() missing argument 'positions' which is not 'None'")
    if sizes is None:
        raise ValueError("mondrians() missing argument 'sizes' which is not 'None'")
    if intensities is None:
        raise ValueError("mondrians() missing argument 'intensities' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    img = np.ones(shape) * intensity_background
    mask = np.zeros(shape)

    n_mondrians = len(positions)
    ints = itertools.cycle(intensities)

    if isinstance(sizes, (float, int)):
        sizes = ((sizes, sizes),) * n_mondrians

    if any(len(lst) != n_mondrians for lst in [positions, sizes]):
        raise Exception("As many positions as sizes required.")

    epositions_px = []
    eshapes = []

    for m in range(n_mondrians):
        try:
            if len(positions[m]) != 2:
                raise ValueError("Position tuples should be (ypos, xpos)")
        except Exception:
            raise ValueError("Position tuples should be (ypos, xpos)")

        ypos, xpos = resolution.lengths_from_visual_angles_ppd(positions[m], ppd[0])
        individual_shapes = resolution.lengths_from_visual_angles_ppd(sizes[m], ppd[0])

        try:
            if len(individual_shapes) == 2:
                depth = 0
                individual_shapes = individual_shapes + [
                    depth,
                ]
            elif len(individual_shapes) == 3:
                depth = sizes[m][2]
            else:
                raise ValueError(
                    "Mondrian size tuples should be (height, width) for "
                    "rectangles or (height, width, depth) for parallelograms"
                )
        except Exception:
            raise ValueError(
                "Mondrian size tuples should be (height, width) for"
                "rectangles or (height, width, depth) for parallelograms"
            )

        if depth < 0:
            xpos += int(depth * ppd[0])
        epositions_px.append(tuple([ypos, xpos]))
        eshapes.append(tuple(individual_shapes))

        # Create parallelogram
        patch = parallelogram(
            visual_size=(sizes[m][0], sizes[m][1] + np.abs(depth)),
            ppd=ppd,
            parallelogram_size=(sizes[m][0], sizes[m][1], depth),
        )

        # Place it into Mondrian mosaic
        yshape, xshape = patch["img"].shape
        if ypos < 0 or xpos < 0:
            raise ValueError("There are no negative position coordinates")
        if (ypos + yshape > shape[0]) or (xpos + xshape > shape[1]):
            raise ValueError("Not all Mondrians fit into the stimulus")
        mask_large = np.zeros(shape)
        mask_large[ypos : ypos + yshape, xpos : xpos + xshape] = patch["parallelogram_mask"]

        img[mask_large == 1] = next(ints)
        mask[mask_large == 1] = m + 1

    stim = {
        "img": img,
        "mondrian_mask": mask.astype(int),
        "ppd": ppd,
        "visual_size": visual_size,
        "shape": shape,
        "positions": tuple(positions),
        "sizes": tuple(sizes),
        "intensities": tuple(intensities),
        "intensity_background": intensity_background,
    }
    return stim


def corrugated_mondrian(
    visual_size=None,
    ppd=None,
    shape=None,
    nrows=None,
    ncols=None,
    depths=0,
    intensities=(0, 1),
    target_indices=(),
    intensity_background=0.5,
    intensity_target=None,
):
    """Corrugated Mondrian

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    depths : Sequence[Number, ... ], Number, or None (default)
        depth of Mondrian parallelograms per row
    intensities : nested tuples
        intensities of mondrians; as many tuples as there are rows and as many
        numbers in each tuple as there are columns
    target_indices : nested tuples
        indices of targets; as many tuples as there are targets with (y, x) indices
    intensity_background : float
        intensity value for background
    intensity_target : float
            target intensity. If None, use values defined in intensities

    Returns
    ------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Adelson, E. H. (1993).
        Perceptual organization and the judgment of brightness.
        Science, 262, 2042-2044.
        https://doi.org/10.1126/science.8266102
    """

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if nrows is None:
        nrows = len(intensities)

    try:
        if ncols is None:
            ncols = len(intensities[0])
    except Exception():
        ncols = nrows

    if isinstance(depths, (float, int)):
        depths = (depths,) * nrows

    if target_indices is None:
        target_indices = ()

    if len(depths) != nrows:
        raise ValueError("Unclear number of rows. Check nrows, intensities and depths.")

    ints = itertools.cycle(tuple(np.array(intensities).flatten()))

    height, width = visual_size
    mdepths_px = resolution.lengths_from_visual_angles_ppd(depths, ppd[0])

    s1 = sum(i for i in mdepths_px if i > 0)
    s2 = sum(-i for i in mdepths_px if i < 0)
    sum_depth = np.maximum(abs(s1), abs(s2))
    r1 = np.cumsum(np.array(mdepths_px)).max()
    r2 = np.cumsum(np.array(mdepths_px)).min()
    red_depth = np.maximum(abs(r1), abs(r2))
    red_depth = np.maximum(red_depth, sum_depth)
    mheight_px, mwidth_px = int(shape[0] / nrows), int((shape[1] - red_depth) / ncols)

    # Initial y coordinates
    yst = 0

    # Calculate initial x coordinates
    xstarts = np.cumsum(np.hstack([0, mdepths_px]))
    temp = np.hstack([mdepths_px, 0])
    temp[temp > 0] = 0.0
    xstarts += temp
    xstarts += np.abs(xstarts.min())

    sizes = []
    poses = []
    intenses = []
    tlist = []
    target_counter = 1
    counter = 0

    for r in range(nrows):
        xst = xstarts[r]
        if depths[r] < 0:
            xst -= int(depths[r] * ppd[0])

        for c in range(ncols):
            if c != ncols - 1:
                msize = (mheight_px / ppd[0], (mwidth_px + 1) / ppd[1], depths[r])
            else:
                msize = (mheight_px / ppd[0], mwidth_px / ppd[1], depths[r])
            mpos = (yst / ppd[0], xst / ppd[1])
            mint = next(ints)

            sizes.append(msize)
            poses.append(mpos)
            intenses.append(mint)

            if (r, c) in target_indices:
                tlist.append(target_counter)

            xst += mwidth_px
            target_counter += 1
            counter += 1
        yst += mheight_px

    stim = mondrian(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        positions=poses,
        sizes=sizes,
        intensities=intenses,
        intensity_background=intensity_background,
    )

    target_mask = np.zeros(shape)
    for t in range(len(tlist)):
        target_mask[stim["mondrian_mask"] == tlist[t]] = t + 1

    stim["target_mask"] = target_mask.astype(int)
    stim["target_indices"] = target_indices

    if intensity_target is not None:
        for t in range(len(tlist)):
            stim["img"] = np.where(target_mask == t + 1, intensity_target, stim["img"])

    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "mondrian": mondrian(**default_params,
                             positions=((0, 0), (8, 4), (1, 6), (4, 4), (5, 1)),
                             sizes=((3, 4, 1), (2, 2, 0), (5, 4, -1), (3, 4, 1), (5, 2, 0)),
                             intensities=np.random.rand(5)),
        "corrugated_mondrian": corrugated_mondrian(**default_params,
                                                   depths=(1, 0, -1, 0),
                                                   intensities=np.random.rand(4, 4),
                                                   target_indices=((1, 1), (3, 1))),
        "corrugated_mondrian2": corrugated_mondrian(**default_params,
                                                    nrows=5,
                                                    ncols=5,
                                                    depths=(1, -1, 0, -1, 1),
                                                    intensities=(0, 1))
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Modelfest dataset from Carney et al (1999) https://doi.org/10.1117/12.348473

This module reproduces all of the stimuli from the Modelfest dataset
prior to the spatiotemporal extension as described in Carney et al (1999)
but normalized between 0 and 1.
More information on Modelfest can be found here:
https://www.visionscience.com/data/modelfest/

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.modelfest.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/modelfest.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.modelfest import *

References
-----------
Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
    ... & Eckstein, M. P. (1999).
    Development of an image/threshold database
    for designing and testing human vision models.
    Proceedings of SPIE, 3644, 542-551.
    https://doi.org/10.1117/12.348473
"""

from pathlib import Path

import numpy as np
import pandas as pd

from stimupy import checkerboards
from stimupy.components import gaussians, lines, shapes
from stimupy.components.edges import gaussian as gaussian_edge
from stimupy.components.waves import bessel
from stimupy.noises.binaries import binary as binary_noise
from stimupy.stimuli.gabors import gabor
from stimupy.stimuli.plaids import gabors as plaid
from stimupy.utils import pad_dict_to_shape, resize_dict, resolution, roll_dict, stack_dicts

__all__ = [
    "GaborPatch1",
    "GaborPatch2",
    "GaborPatch3",
    "GaborPatch4",
    "GaborPatch5",
    "GaborPatch6",
    "GaborPatch7",
    "GaborPatch8",
    "GaborPatch9",
    "GaborPatch10",
    "GaborPatch11",
    "GaborPatch12",
    "GaborPatch13",
    "GaborPatch14",
    "ElongatedGabor15",
    "ElongatedGabor16",
    "ElongatedGabor17",
    "Baguette18",
    "Baguette19",
    "Baguette20",
    "Baguette21",
    "Subthreshold22",
    "Subthreshold23",
    "Subthreshold24",
    "Subthreshold25",
    "Gaussians26",
    "Gaussians27",
    "Gaussians28",
    "Gaussians29",
    "Edge30",
    "Line31",
    "Dipole32",
    "GaborString33",
    "GaborString34",
    "Noise35",
    "Orientation36",
    "Orientation37",
    "Plaids38",
    "Plaids39",
    "Disk40",
    "Bessel41",
    "Checkerboard42",
    "NaturalScene43",
]

# Default values:
PPD = 120  # pixel size of 0.5 arcmin
PPD2 = 60  # pixel size of 1 arcmin (pixel replication)
VISUAL_SIZE = resolution.visual_size_from_shape_ppd(shape=256, ppd=PPD)

mean_lum = 0.5

df = pd.read_csv(Path(__file__).parents[0] / "modelfest_data.csv", header=None)
participants = df[0]


def gen_all(ppd=PPD, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating modelfest.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def GaborPatch1(ppd=PPD):
    """GaborPatch1 with fixed size in degrees, Carney et al (1999)
    Frequency: 1.12 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 1.12,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 1
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch2(ppd=PPD):
    """GaborPatch2 with fixed size in degrees, Carney et al (1999)
    Frequency: 2 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 2,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 5
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch3(ppd=PPD):
    """GaborPatch3 with fixed size in degrees, Carney et al (1999)
    Frequency: 2.83 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 2.83,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 9
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch4(ppd=PPD):
    """GaborPatch4 with fixed size in degrees, Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 13
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch5(ppd=PPD):
    """GaborPatch5 with fixed size in degrees, Carney et al (1999)
    Frequency: 5.66 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 5.66,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 17
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch6(ppd=PPD):
    """GaborPatch6 with fixed size in degrees, Carney et al (1999)
    Frequency: 8 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 8,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 21
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch7(ppd=PPD):
    """GaborPatch7 with fixed size in degrees, Carney et al (1999)
    Frequency: 11.3 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 11.3,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 25
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch8(ppd=PPD):
    """GaborPatch8 with fixed size in degrees, Carney et al (1999)
    Frequency: 16 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 16,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 29
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch9(ppd=PPD):
    """GaborPatch9 with fixed size in degrees, Carney et al (1999)
    Frequency: 22.6 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 22.6,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 33
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch10(ppd=PPD):
    """GaborPatch10 with fixed size in degrees, Carney et al (1999)
    Frequency: 30 cpd
    Gaussian window: sx=sy=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 30,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 37
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch11(ppd=PPD):
    """GaborPatch11 with fixed size in cycles (~1 octave), Carney et al (1999)
    Frequency: 2 cpd
    Gaussian window: sx=sy=0.28 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 2,
        "sigma": 0.28,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 41
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch12(ppd=PPD):
    """GaborPatch12 with fixed size in cycles (~1 octave), Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sx=sy=0.14 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.14,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 45
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch13(ppd=PPD):
    """GaborPatch13 with fixed size in cycles (~1 octave), Carney et al (1999)
    Frequency: 8 cpd
    Gaussian window: sx=sy=0.07 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 8,
        "sigma": 0.07,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 49
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborPatch14(ppd=PPD):
    """GaborPatch14 with fixed size in cycles (~1 octave), Carney et al (1999)
    Frequency: 16 cpd
    Gaussian window: sx=sy=0.035 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 16,
        "sigma": 0.035,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 53
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def ElongatedGabor15(ppd=PPD):
    """ElongatedGabor15, Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sy=0.28 (~0.5octave); sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": (0.28, 0.5),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 57
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def ElongatedGabor16(ppd=PPD):
    """ElongatedGabor16, Carney et al (1999)
    Frequency: 8 cpd
    Gaussian window: sy=0.14 (~0.5octave); sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 8,
        "sigma": (0.14, 0.5),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 61
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def ElongatedGabor17(ppd=PPD):
    """ElongatedGabor17, Carney et al (1999)
    Frequency: 16 cpd
    Gaussian window: sy=0.07 (~0.5octave); sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 16,
        "sigma": (0.07, 0.5),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 65
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Baguette18(ppd=PPD):
    """Baguette18 (elongated Gabors; varying aspect ratios), Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sy=0.14 (~1octave); sx=0.28 deg (~2octaves)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": (0.14, 0.28),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 69
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Baguette19(ppd=PPD):
    """Baguette19 (elongated Gabors; varying aspect ratios), Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sy=0.14 (~1octave); sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": (0.14, 0.5),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 73
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Baguette20(ppd=PPD):
    """Baguette20 (elongated Gabors; varying aspect ratios), Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sy=0.28 (~2octave); sx=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": (0.28, 0.14),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 77
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Baguette21(ppd=PPD):
    """Baguette20 (elongated Gabors; varying aspect ratios), Carney et al (1999)
    Frequency: 4 cpd
    Gaussian window: sy=0.5 deg; sx=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": (0.5, 0.14),
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 81
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Subthreshold22(ppd=PPD):
    """Subthreshold22 (compound Gabor), Carney et al (1999)
    Frequencies: 2 cpd & 2*sqrt(2) cpd
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim1 = gabor(**params, frequency=2)
    stim2 = gabor(**params, frequency=2 * np.sqrt(2))

    stim1["img"] = stim1["img"] / 2 + stim2["img"] / 2
    stim1["grating_mask2"] = stim2["grating_mask"]
    stim1["frequency2"] = stim2["frequency"]
    stim1["phase_width2"] = stim2["phase_width"]
    stim1["n_phases2"] = stim2["n_phases"]

    v = 85
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim1, "experimental_data": experimental_data}


def Subthreshold23(ppd=PPD):
    """Subthreshold23 (compound Gabor), Carney et al (1999)
    Frequencies: 2 cpd & 4 cpd
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim1 = gabor(**params, frequency=2)
    stim2 = gabor(**params, frequency=4)

    stim1["img"] = stim1["img"] / 2 + stim2["img"] / 2
    stim1["grating_mask2"] = stim2["grating_mask"]
    stim1["frequency2"] = stim2["frequency"]
    stim1["phase_width2"] = stim2["phase_width"]
    stim1["n_phases2"] = stim2["n_phases"]

    v = 89
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim1, "experimental_data": experimental_data}


def Subthreshold24(ppd=PPD):
    """Subthreshold24 (compound Gabor), Carney et al (1999)
    Frequencies: 4 cpd & 4*sqrt(2) cpd
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim1 = gabor(**params, frequency=4)
    stim2 = gabor(**params, frequency=4 * np.sqrt(2))

    stim1["img"] = stim1["img"] / 2 + stim2["img"] / 2
    stim1["grating_mask2"] = stim2["grating_mask"]
    stim1["frequency2"] = stim2["frequency"]
    stim1["phase_width2"] = stim2["phase_width"]
    stim1["n_phases2"] = stim2["n_phases"]

    v = 93
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim1, "experimental_data": experimental_data}


def Subthreshold25(ppd=PPD):
    """Subthreshold24 (compound Gabor), Carney et al (1999)
    Frequencies: 4 cpd & 8 cpd
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 0.5,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim1 = gabor(**params, frequency=4)
    stim2 = gabor(**params, frequency=8)

    stim1["img"] = stim1["img"] / 2 + stim2["img"] / 2
    stim1["grating_mask2"] = stim2["grating_mask"]
    stim1["frequency2"] = stim2["frequency"]
    stim1["phase_width2"] = stim2["phase_width"]
    stim1["n_phases2"] = stim2["n_phases"]

    v = 97
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim1, "experimental_data": experimental_data}


def Gaussians26(ppd=PPD):
    """Gaussians26, Carney et al (1999)
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 0.5,
        "intensity_max": 1,
        "origin": "center",
    }

    stim = gaussians.gaussian(**params)
    stim["img"] = stim["img"] / 2 + 0.5

    v = 101
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Gaussians27(ppd=PPD):
    """Gaussians27, Carney et al (1999)
    Gaussian window: sy=sx=8.43 arcmin

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 8.43 / 60,
        "intensity_max": 1,
        "origin": "center",
    }

    stim = gaussians.gaussian(**params)
    stim["img"] = stim["img"] / 2 + 0.5

    v = 105
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Gaussians28(ppd=PPD):
    """Gaussians28, Carney et al (1999)
    Gaussian window: sy=sx=2.106 arcmin

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 2.106 / 60,
        "intensity_max": 1,
        "origin": "center",
    }

    stim = gaussians.gaussian(**params)
    stim["img"] = stim["img"] / 2 + 0.5

    v = 109
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Gaussians29(ppd=PPD):
    """Gaussians29, Carney et al (1999)
    Gaussian window: sy=sx=1.05 arcmin

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "sigma": 1.05 / 60,
        "intensity_max": 1,
        "origin": "center",
    }

    stim = gaussians.gaussian(**params)
    stim["img"] = stim["img"] / 2 + 0.5

    v = 113
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Edge30(ppd=PPD):
    """Edge30, Carney et al (1999)
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = gaussian_edge(visual_size=VISUAL_SIZE, ppd=ppd, rotation=-90, sigma=0.5)

    v = 117
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Line31(ppd=PPD):
    """Line31: line x Gaussian, Carney et al (1999)
    Line width: 0.5 arcmin
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = lines.line(
        visual_size=VISUAL_SIZE,
        ppd=ppd,
        line_length=VISUAL_SIZE.width,
        line_width=resolution.visual_angle_from_length_ppd(length=0.5, ppd=PPD2),
        rotation=90,
        intensity_background=0.5,
    )
    window = gaussians.gaussian(visual_size=VISUAL_SIZE, ppd=ppd, sigma=0.5)

    img = (stim["img"] - 0.5) * window["img"] + 0.5
    stim["img"] = img
    stim["sigma"] = window["sigma"]

    v = 121
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Dipole32(ppd=PPD):
    """Dipole32: lines x Gaussian, Carney et al (1999)
    Line width: 3 px =
    Separation: 1 px =
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = lines.dipole(
        visual_size=VISUAL_SIZE,
        ppd=ppd,
        line_length=VISUAL_SIZE.width,
        line_width=resolution.visual_angle_from_length_ppd(length=1, ppd=PPD),
        line_gap=resolution.visual_angle_from_length_ppd(length=2, ppd=PPD),
        rotation=270,
    )
    stim = roll_dict(stim, -1, axes=0)
    window = gaussians.gaussian(visual_size=VISUAL_SIZE, ppd=ppd, sigma=0.5)

    img = (stim["img"] - 0.5) * window["img"] + 0.5
    stim["img"] = img
    stim["sigma"] = window["sigma"]

    v = 125
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborString33(ppd=PPD):
    """GaborString33 - 5 collinear in-phase Gabors, Carney et al (1999)
    Frequency: 8 cpd
    Gaussian window of individual Gabors: sx=sy=0.07 deg (~1octave)
    Separation: 5sx

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE.width / 6,
        "ppd": ppd,
        "frequency": 8,
        "sigma": 0.07,
        "rotation": -90,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    # Stack collinear Gabors horizontally and pad
    stimc = stack_dicts(stim, stim)
    stimc = stack_dicts(stimc, stimc)
    stimc = stack_dicts(stimc, stim)
    stimc = pad_dict_to_shape(stimc, (256, 256), pad_value=0.5)

    stim["img"] = stimc["img"]
    stim["grating_mask"] = stimc["grating_mask"]
    stim["visual_size"] = stimc["visual_size"]
    stim["shape"] = stimc["shape"]
    stim = roll_dict(stim, (-1, -1), axes=(0, 1))

    v = 129
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def GaborString34(ppd=PPD):
    """GaborString33 - 5 collinear out-of-phase Gabors, Carney et al (1999)
    Frequency: 8 cpd
    Gaussian window of individual Gabors: sx=sy= deg (~1octave)
    Separation: 5sx

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE.width / 6,
        "ppd": ppd,
        "frequency": 8,
        "sigma": 0.07,
        "rotation": -90,
        "origin": "center",
    }

    stim1 = gabor(**params, phase_shift=-90)
    stim2 = gabor(**params, phase_shift=90)

    # Stack collinear Gabors horizontally and pad
    stimc = stack_dicts(stim2, stim1)
    stimc = stack_dicts(stimc, stim2)
    stimc = stack_dicts(stimc, stim1)
    stimc = stack_dicts(stimc, stim2)
    stimc = pad_dict_to_shape(stimc, (256, 256), pad_value=0.5)

    stim1["img"] = stimc["img"]
    stim1["grating_mask"] = stimc["grating_mask"]
    stim1["visual_size"] = stimc["visual_size"]
    stim1["shape"] = stimc["shape"]
    stim1["phase_shift2"] = stim2["phase_shift"]
    stim1 = roll_dict(stim1, (-1, -1), axes=(0, 1))

    v = 133
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim1, "experimental_data": experimental_data}


def Noise35_random(ppd=PPD):
    """Noise35 - binary noise x Gaussian, Carney et al (1999)
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = binary_noise(visual_size=VISUAL_SIZE, ppd=ppd / 2, rms_contrast=1)
    stim = resize_dict(stim, (2, 2))
    window = gaussians.gaussian(visual_size=VISUAL_SIZE, ppd=ppd, sigma=0.5)

    img = stim["img"] * window["img"]
    stim["img"] = img / 2 + 0.5
    stim["mask"] = np.zeros(img.shape).astype(int)
    return stim


def Noise35(ppd=PPD):
    """Noise35 - binary noise x Gaussian, Carney et al (1999)
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """
    # Read natural image from Modelfest
    img = read_tif(Path(__file__).parents[0] / "modelfest_noise.tif")
    img = img / 255

    stim = {
        "img": img,
        "visual_size": VISUAL_SIZE,
        "shape": img.shape,
        "ppd": PPD,
        "intensity_range": (img.min(), img.max()),
    }

    v = 169
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    mask = np.zeros(img.shape).astype(int)
    return {**stim, "mask": mask, "experimental_data": experimental_data}


def Orientation36(ppd=PPD):
    """Orientation36 - oriented Gabor, Carney et al (1999)
    Frequency: 4 cpd
    Rotation: 45 deg
    Gaussian window: sx=sy=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.14,
        "rotation": -45,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 141
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Orientation37(ppd=PPD):
    """Orientation37 - oriented Gabor, Carney et al (1999)
    Frequency: 4 cpd
    Rotation: 0 deg
    Gaussian window: sx=sy=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.14,
        "rotation": 0,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = gabor(**params)

    v = 145
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Plaids38(ppd=PPD):
    """Plaids38 - compound Gabors, Carney et al (1999)
    Frequency: 4 cpd
    Rotation: 0 & 90 deg
    Gaussian window: sx=sy=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.14,
        "phase_shift": 90,
        "origin": "center",
        "round_phase_width": False,
    }

    stim = plaid(
        gabor_parameters1={**params, "rotation": 0},
        gabor_parameters2={**params, "rotation": -90},
    )

    v = 149
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Plaids39(ppd=PPD):
    """Plaids39 - compound Gabors, Carney et al (1999)
    Frequency: 4 cpd
    Rotation: 45 & 90 deg
    Gaussian window: sx=sy=0.14 deg (~1octave)

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 4,
        "sigma": 0.14,
        "phase_shift": 90,
        "origin": "center",
    }

    stim = plaid(
        gabor_parameters1={**params, "rotation": -45},
        gabor_parameters2={**params, "rotation": -90},
    )

    v = 153
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Disk40(ppd=PPD):
    """Disk40 - Carney et al (1999)
    Diameter: 0.25 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = shapes.disc(
        visual_size=VISUAL_SIZE, ppd=ppd, radius=0.1249, origin="center", intensity_background=0.5
    )
    stim = roll_dict(stim, (-2, -2), axes=(0, 1))

    v = 157
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Bessel41(ppd=PPD):
    """Bessel41 - Bessel x Gaussian, Carney et al (1999)
    Frequency: 4cpd
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    stim = bessel(visual_size=VISUAL_SIZE, ppd=ppd, frequency=4, origin="center")
    window = gaussians.gaussian(visual_size=VISUAL_SIZE, ppd=ppd, sigma=0.5)
    stim["sigma"] = window["sigma"]

    # Apply Gaussian windows to Bessel
    stim["img"] = (stim["img"] - stim["img"].mean()) * window["img"]

    # Set "background" intensity to 0.5 and make sure that max intensity = 1
    stim["img"] = stim["img"] / stim["img"].max() / 2 + 0.5

    v = 161
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def Checkerboard42(ppd=PPD):
    """Checkerboard42 - Checkerboard x Gaussian, Carney et al (1999)
    Frequency: 4cpd fundamental
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """
    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "frequency": 2.81,
        "rotation": -45,
        "round_phase_width": False,
        "intensity_checks": (1.0, 0.0),
    }

    stim = checkerboards.checkerboard(**params)
    window = gaussians.gaussian(visual_size=VISUAL_SIZE, ppd=ppd, sigma=0.5)

    img = (stim["img"] - 0.5) * window["img"] + 0.5
    stim["img"] = img
    stim["sigma"] = window["sigma"]

    v = 165
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    return {**stim, "experimental_data": experimental_data}


def NaturalScene43(ppd=PPD):
    """NaturalScene43 - Natural image x Gaussian, Carney et al (1999)
    Gaussian window: sy=sx=0.5 deg

    Parameters
    ----------
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal]

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and additional keys containing
        stimulus parameters

    References
    ----------
    Carney, T., Klein, S. A., Tyler, C. W., Silverstein, A. D., Beutter, B., Levi, D.,
        ... & Eckstein, M. P. (1999).
        Development of an image/threshold database
        for designing and testing human vision models.
        Proceedings of SPIE, 3644, 542-551.
        https://doi.org/10.1117/12.348473
    """

    # Read natural image from Modelfest
    img = read_tif(Path(__file__).parents[0] / "modelfest_natural_scene.tif")
    img /= 255

    stim = {
        "img": img,
        "visual_size": VISUAL_SIZE,
        "shape": img.shape,
        "ppd": PPD,
        "intensity_range": (img.min(), img.max()),
    }

    v = 169
    experimental_data = {
        "participants": participants,
        "thresholds1": df[v],
        "thresholds2": df[v + 1],
        "thresholds3": df[v + 2],
        "thresholds4": df[v + 3],
    }
    mask = np.zeros(img.shape).astype(int)
    return {**stim, "mask": mask, "experimental_data": experimental_data}


def read_tif(filename):
    from PIL import Image

    img = np.array(Image.open(filename)).astype(float)
    return img


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=False, units="visual_size")



"""Stimuli from White & White (1985)
https://doi.org/10.1016/0042-6989(85)90049-5

This module reproduces most of the stimuli used by White & White (1985) as
they were described in that paper.

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.white1985.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/white1985.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.white1985 import *

References
-----------
White, M. & White, T. (1985).
    Counterphase lightness induction.
    Vision Research, 25 (9), 1331-1335.
    https://doi.org/10.1016/0042-6989(85)90049-5
"""

import copy
import warnings

import numpy as np

from stimupy import gratings

__all__ = [
    "wide_0phase",
    "wide_36phase",
    "wide_72phase",
    "wide_108phase",
    "wide_144phase",
    "wide_180phase",
    "square_0phase",
    "square_36phase",
    "square_72phase",
    "square_108phase",
    "square_144phase",
    "square_180phase",
]

PPD = 50
BAR_WIDTH = 1 / 3.5 / 2
v1, v2, v3 = 0.0, 0.5, 1.0
START_PHASE1 = -180
START_PHASE2 = 0

COMMON_PARAMS = {
    "intensity_bars": (v3, v1),
    "intensity_target": v2,
    "distance_metric": "vertical",
}


def gen_all(ppd=PPD, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating white1981.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def resolve_bar_width(bar_width=BAR_WIDTH, ppd=PPD):
    bar_width_old = copy.deepcopy(bar_width)
    bar_width = np.round(bar_width * ppd) / ppd

    if bar_width_old != bar_width:
        warnings.warn(
            f"Rounding bar_width because of ppd; {bar_width_old} -> {bar_width}. "
            "This will also effect the stimulus size."
        )
    return bar_width


def wide_0phase(ppd=PPD):
    """A square-wave grating with four bars that are in phase as shown in
    White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: ~3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 0,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def wide_36phase(ppd=PPD):
    """A square-wave grating with four bars that are 36 deg out-of-phase as
    shown in White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 36,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def wide_72phase(ppd=PPD):
    """A square-wave grating with four bars that are 72 deg out-of-phase as
    shown in White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 72,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def wide_108phase(ppd=PPD):
    """A square-wave grating with four bars that are 108 deg out-of-phase as
    shown in White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 108,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def wide_144phase(ppd=PPD):
    """A square-wave grating with four bars that are 144 deg out-of-phase as
    shown in White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 144,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def wide_180phase(ppd=PPD):
    """A square-wave grating with four bars that are 180 deg out-of-phase as
    shown in White & White (1985), Fig. 2
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.75 deg (originally 0.15 x 0.75 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, 0.75),
        "target_phase_shift": START_PHASE1 + 180,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_0phase(ppd=PPD):
    """A square-wave grating with four squares that are in phase as shown in
    White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 0,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_36phase(ppd=PPD):
    """A square-wave grating with four squares that are 36 deg out-of-phase as
    shown in White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 36,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_72phase(ppd=PPD):
    """A square-wave grating with four squares that are 72 deg out-of-phase as
    shown in White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 72,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_108phase(ppd=PPD):
    """A square-wave grating with four squares that are 108 deg out-of-phase as
    shown in White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 108,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_144phase(ppd=PPD):
    """A square-wave grating with four squares that are 144 deg out-of-phase as
    shown in White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 144,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def square_180phase(ppd=PPD):
    """A square-wave grating with four squares that are 180 deg out-of-phase as
    shown in White & White (1985), Fig. 3
    Stimulus size: 3.5 x 3.5 deg
    Target bars: 0.14 x 0.14 deg (originally 0.15 x 0.15 deg)
    Grating frequency: 3.5 cpd

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. & White, T. (1985).
        Counterphase lightness induction.
        Vision Research, 25 (9), 1331-1335.
        https://doi.org/10.1016/0042-6989(85)90049-5
    """

    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": 25 * bar_width,
        "ppd": ppd,
        "bar_width": bar_width,
        "target_size": (bar_width * 9, bar_width),
        "target_phase_shift": START_PHASE2 + 180,
        **COMMON_PARAMS,
    }

    stim = gratings.phase_shifted(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=False)



import numpy as np

from stimupy.utils import resolution
from stimupy.utils.filters import convolve

__all__ = [
    "wedding_cake",
]


def wedding_cake(
    visual_size=None,
    ppd=None,
    shape=None,
    L_size=None,
    target_height=None,
    target_indices1=None,
    target_indices2=None,
    intensity_bars=(0.0, 1.0),
    intensity_target=0.5,
):
    """Wedding cake stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    L_size : (float, float, float)
        size of individual jags (height, width, thickness) in degree visual angle
    target_height : float
        height of targets in degree visual angle
    target_indices1 : nested tuples
        target indices with intensity1-value; as many tuples as there are targets
        each with (y, x) indices
    target_indices2 : nested tuples
        target indices with intensity2-value; as many tuples as there are targets
        each with (y, x) indices
    intensity_bars : (float, float)
        intensity values of the bars
    intensity_target : float
        intensity value of targets

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Clifford, C. W. G., & Spehar, B. (2003).
        Using colour to disambiguate contrast and assimilation in White's effect.
        Journal of Vision, 3, 294a.
        https://doi.org/10.1167/3.9.294
    """
    if L_size is None:
        raise ValueError("wedding_cake() missing argument 'L_size' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd_ = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")
    if visual_size[1] < visual_size[0] * 0.75 or visual_size[1] > visual_size[0] * 1.25:
        raise ValueError(
            "wedding_cake not available for images with very different heights+widths"
        )

    nY, nX = shape
    Ly, Lx, Lw = resolution.lengths_from_visual_angles_ppd(L_size, np.unique(ppd))
    Lyh, Lxh = int(Ly / 2) + 1, int(Lx / 2) + 1
    if Lw > Ly / 2:
        raise ValueError("L-thickness should not exceed L_height / 2")

    # Create L-shaped patch
    L_patch = np.zeros([Ly, Lx])
    L_patch[0:Lw, 0:Lx] = 1
    L_patch[0:Ly, Lx - Lw : :] = 1

    # We initially create a larger image array to avoid boundary problems
    nY, nX = nY * 2, nX * 2

    # Create grid-like array to create wedding cake pattern
    array1 = np.zeros([nY, nX])
    ys, xs = np.arange(0, nY, Ly - Lw), np.arange(0, nX, Lx)
    n = np.minimum(len(ys), len(xs))
    array1[ys[0:n], xs[0:n]] = np.arange(1, n + 1)

    array2 = np.zeros([nY, nX])
    ys, xs = np.arange(0, nY, Lw * 2), np.arange(0, nX, Lw * 2)
    n = np.minimum(len(ys), len(xs))
    array2[ys[0:n], xs[0:n]] = np.arange(1, n + 1)
    array2 = np.rot90(array2)
    amax = int((array2.max() + 1) / 2)

    array3 = convolve(array1, array2, "same")

    if target_indices1 is not None and target_height is not None:
        # Create target patch2
        theight = resolution.lengths_from_visual_angles_ppd(target_height, np.unique(ppd))
        tpatch1 = np.zeros(L_patch.shape)
        tpatch1[int(Ly / 2 - theight / 2) : int(Ly / 2 + theight / 2), Lx - Lw : :] = (
            -intensity_bars[1] + intensity_target
        )

        array_t1 = np.zeros(array3.shape)
        for ty, tx in target_indices1:
            arr1 = np.copy(array1)
            arr2 = np.copy(array2)
            arr1[arr1 != ty + 2] = 0
            arr2[arr2 != tx + amax] = 0
            array_t1 += convolve(arr1, arr2, "same")
        array_t1[array_t1 < 1] = 0
        array_t1[array_t1 > 1] = 1
        t1 = np.round(convolve(array_t1, tpatch1, "same"), 5)
        t1 = t1[Lyh : Lyh + int(nY / 2), Lxh : Lxh + int(nX / 2)]

    else:
        t1 = np.zeros([int(nY / 2), int(nX / 2)])

    if target_indices2 is not None and target_height is not None:
        # Create target patch2
        theight = resolution.lengths_from_visual_angles_ppd(target_height, np.unique(ppd))
        tpatch2 = np.zeros(L_patch.shape)
        tpatch2[int(Ly / 2 - theight / 2) : int(Ly / 2 + theight / 2), Lx - Lw : :] = (
            -intensity_bars[0] + intensity_target
        )

        array_t2 = np.zeros(array3.shape)
        for ty, tx in target_indices2:
            arr1 = np.copy(array1)
            arr2 = np.copy(array2)
            arr1[arr1 != ty + 2] = 0
            arr2[arr2 != tx + amax] = 0
            array_t2 += convolve(arr1, arr2, "same")
        array_t2[array_t2 < 1] = 0
        array_t2[array_t2 > 1] = 1
        t2 = np.round(convolve(array_t2, tpatch2, "same"), 5)
        t2 = t2[Lyh - Lw : Lyh + int(nY / 2) - Lw, Lxh + Lw : Lxh + int(nX / 2) + Lw]

    else:
        t2 = np.zeros([int(nY / 2), int(nX / 2)])

    # Create wedding cake pattern
    imgt = convolve(array3, L_patch, "same")
    imgt = np.round(imgt)
    img = np.copy(imgt)
    img[imgt > 1] = intensity_bars[1]
    img[imgt < 1] = intensity_bars[0]
    img = img[Lyh : Lyh + int(nY / 2), Lxh : Lxh + int(nX / 2)]

    # Create target mask
    mask = np.abs(t1 * 1.0 / intensity_target) + np.abs(t2 * 2.0 / intensity_target)
    mask = np.round(mask).astype(int)

    # Add targets
    img = img + t1
    img = img + t2

    stim = {
        "img": img,
        "target_mask": mask.astype(int),
        "shape": shape,
        "visual_size": visual_size,
        "ppd": ppd,
        "L_size": L_size,
        "target_height": target_height,
        "intensity_bars": intensity_bars,
        "intensity_target": intensity_target,
        "target_indices1": target_indices1,
        "target_indices2": target_indices2,
    }

    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 15,
        "ppd": 10,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "wedding_cake": wedding_cake(**default_params,
                                     L_size=(4, 3, 1),
                                     target_height=1,
                                     target_indices1=((2, 2), (2, 1)),
                                     target_indices2=((2, -1), (2, 0)),
                                     )
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



from stimupy.components import lines
from stimupy.utils import resolution, stack_dicts

__all__ = [
    "ponzo",
]


def ponzo(
    visual_size=None,
    ppd=None,
    shape=None,
    outer_lines_length=None,
    outer_lines_width=0,
    outer_lines_angle=15,
    target_lines_length=None,
    target_lines_width=0,
    target_distance=None,
    intensity_outer_lines=1.0,
    intensity_target_lines=0.5,
    intensity_background=0.0,
):
    """Ponzo's (1910) illusion

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    outer_lines_length : Number
        length of outer lines in degrees visual angle
    outer_lines_width : Number
        line width of outer lines in degrees visual angle
        if 0 (default), set line width to 1 px
    outer_lines_angle : Number
        angle of outer lines in degrees. Must be between -45 and 45 degrees.
    target_lines_length : Number
        length of target lines in degrees visual angle
    target_lines_width :
        line width of target lines in degrees visual angle
        if 0 (default), set line width to 1 px
    target_distance : Number
        distance between target lines in degrees visual angle
    intensity_outer_lines : Number or (Number, Number)
        intensity value(s) of outer lines
    intensity_target_lines : Number or (Number, Number)
        intensity value(s) of target lines
    intensity_background : Number
        intensity value of background

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Ponzo, M. (1910).
        Intorno ad alcune illusioni nel campo delle sensazioni tattili,
        sull'illusione di Aristotele e fenomeni analoghi.
        Wilhelm Engelmann.
    """
    if outer_lines_length is None:
        raise ValueError("ponzo() missing argument 'outer_lines_length' which is not 'None'")
    if target_lines_length is None:
        raise ValueError("ponzo() missing argument 'target_lines_length' which is not 'None'")
    if target_distance is None:
        raise ValueError("ponzo() missing argument 'target_distance' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    if outer_lines_angle < -45 or outer_lines_angle > 45:
        raise ValueError("outer_lines_angle should be between -45 and 45 deg")
    if isinstance(intensity_outer_lines, (float, int)):
        intensity_outer_lines = (intensity_outer_lines, intensity_outer_lines)
    if isinstance(intensity_target_lines, (float, int)):
        intensity_target_lines = (intensity_target_lines, intensity_target_lines)

    line1 = lines.line(
        ppd=ppd,
        shape=(shape[0], shape[1] / 2),
        line_position=None,
        line_length=outer_lines_length,
        line_width=outer_lines_width,
        rotation=-outer_lines_angle,
        intensity_line=intensity_outer_lines[0] - intensity_background,
        intensity_background=0,
        origin="center",
    )

    line2 = lines.line(
        ppd=ppd,
        shape=(shape[0], shape[1] / 2),
        line_position=None,
        line_length=outer_lines_length,
        line_width=outer_lines_width,
        rotation=outer_lines_angle,
        intensity_line=intensity_outer_lines[1] - intensity_background,
        intensity_background=0,
        origin="center",
    )

    line_position1 = (-target_distance / 2, -target_lines_length / 2)
    line_position2 = (target_distance / 2, -target_lines_length / 2)

    line3 = lines.line(
        ppd=ppd,
        shape=shape,
        line_position=line_position1,
        line_length=target_lines_length,
        line_width=target_lines_width,
        rotation=90,
        intensity_line=intensity_target_lines[0] - intensity_background,
        intensity_background=0,
        origin="center",
    )

    line4 = lines.line(
        ppd=ppd,
        shape=shape,
        line_position=line_position2,
        line_length=target_lines_length,
        line_width=target_lines_width,
        rotation=90,
        intensity_line=intensity_target_lines[1] - intensity_background,
        intensity_background=0,
        origin="center",
    )

    line1 = stack_dicts(line1, line2)
    line1["img"] += line3["img"] + line4["img"] + intensity_background
    line1["line_mask"] += line3["line_mask"] * 3 + line4["line_mask"] * 4
    line1["target_mask"] = line3["line_mask"] + line4["line_mask"] * 2

    stim = {}
    stim["img"] = line1["img"]
    stim["line_mask"] = line1["line_mask"]
    stim["target_mask"] = line1["target_mask"]
    stim["visual_size"] = line1["visual_size"]
    stim["shape"] = line1["shape"]
    stim["ppd"] = line1["ppd"]
    stim["outer_lines_length"] = outer_lines_length
    stim["outer_lines_width"] = outer_lines_width
    stim["outer_lines_angle"] = outer_lines_angle
    stim["target_lines_length"] = target_lines_length
    stim["target_lines_width"] = target_lines_width
    stim["target_distance"] = target_distance
    stim["intensity_outer_lines"] = intensity_outer_lines
    stim["intensity_target_lines"] = intensity_target_lines
    stim["intensity_background"] = intensity_background
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 20,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "ponzo": ponzo(**default_params, outer_lines_angle=10, outer_lines_length=8, target_lines_length=3, target_distance=5)
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=True, save=None)
